# Generated by Django 5.1.7 on 2025-04-10 10:10

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ResilientLogEntry',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('is_sent', models.BooleanField(default=False, verbose_name='is sent')),
                ('level', models.IntegerField(default=0, verbose_name='level')),
                ('message', models.JSONField(verbose_name='message')),
                ('context', models.JSONField(null=True, verbose_name='context')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
            ],
            options={
                'verbose_name_plural': 'resilient log entries',
            },
        ),
    ]



from datetime import timedelta, datetime
from django.db import transaction
from resilient_logger.abstract_log_facade import AbstractLogFacade
from resilient_logger.models import ResilientLogEntry
from typing import Any, Generator, Self, Union, override

class ResilientLogFacade(AbstractLogFacade):
    _log: ResilientLogEntry

    def __init__(self, log: ResilientLogEntry):
        self._log = log

    @override
    def get_id(self) -> Union[str, int]:
        return self._log.id

    @override
    def get_level(self) -> int:
        return self._log.level
    
    @override
    def get_message(self) -> Any:
        return self._log.message
    
    @override
    def get_context(self) -> Any:
        return self._log.context
    
    @override
    def is_sent(self) -> bool:
        return self._log.is_sent

    @override
    def mark_sent(self) -> None:
        self._log.is_sent = True
        self._log.save(update_fields=["is_sent"])

    @override
    @classmethod
    @transaction.atomic
    def create(cls, level: int, message: Any, context: Any) -> Self:
        entry = ResilientLogEntry.objects.create(level=level, message=message, context=context)
        return cls(entry)
    
    @override
    @classmethod
    @transaction.atomic
    def get_unsent_entries(cls, chunk_size: int) -> Generator[AbstractLogFacade, None, None]:
        entries = (
            ResilientLogEntry
                .objects
                .filter(is_sent=False)
                .order_by("created_at")
                .iterator(chunk_size=chunk_size)
        )

        for entry in entries:
            yield ResilientLogFacade(entry)

    @override
    @classmethod
    @transaction.atomic
    def clear_sent_entries(cls, days_to_keep: int = 30) -> None:
        ResilientLogEntry.objects.filter(
            is_sent=True,
            created_at__lte=(datetime.now() - timedelta(days=days_to_keep)),
        ).select_for_update().delete()


import logging
from collections.abc import Callable
from functools import cached_property
from resilient_logger.abstract_log_facade import AbstractLogFacade
from resilient_logger.abstract_submitter import AbstractSubmitter
from typing import Dict, Optional
from resilient_logger.utils import dynamic_class, get_log_record_extra

logger = logging.getLogger(__name__)

class ResilientLogHandler(logging.Handler):
    _submitter_construct: Optional[Callable[[], AbstractSubmitter]]

    def __init__(self, submitter: Dict, log_facade: Dict, level: int = logging.NOTSET):
        super().__init__(level)

        def submitter_construct():
            """
            Submitter rely on Django's DB models and cannot be instantiated during init since Django app registry is not ready by then.
            Work around this by doing lazy initialization when the submitter is used first time.
            """

            submitter_args = submitter.copy()
            submitter_class = submitter_args.pop('class')

            log_facade_args = log_facade.copy()
            log_facade_class = log_facade_args.pop('class')

            Submitter = dynamic_class(AbstractSubmitter, submitter_class)
            LogFacade = dynamic_class(AbstractLogFacade, log_facade_class)

            instance = Submitter(**submitter_args, log_facade=LogFacade)
            return instance

        self._submitter_construct = submitter_construct

    def emit(self, record: logging.LogRecord):
        self._submitter.submit(record.levelno, record.getMessage(), get_log_record_extra(record))

    @cached_property
    def _submitter(self) -> AbstractSubmitter:
        if self._submitter_construct is None:
            raise Exception("Unable to proceed without lazy constructor")
        
        instance = self._submitter_construct()
        self._submitter_construct = None
        return instance
    



from abc import abstractmethod
from django.db import transaction

import logging
import os

from typing import Any, Dict, List, Optional, Generator, Type, TypedDict

from resilient_logger.abstract_log_facade import AbstractLogFacade
from apscheduler.schedulers.background import BackgroundScheduler

logger = logging.getLogger(__name__)

class IntervalOptions(TypedDict):
    """
    Defines internal format for apscheduler trigger definitions (trigger, trigger_args).
    See: https://apscheduler.readthedocs.io/en/3.x/modules/schedulers/base.html#apscheduler.schedulers.base.BaseScheduler.add_job
    """
    trigger: str
    args: Dict[str, Any]

_default_submit_unsent_entries_trigger = IntervalOptions({
    'trigger': 'interval',
    'args': { 'minutes': 15 }
})

_default_clear_sent_entries_trigger = IntervalOptions({
    'trigger': 'cron',
    'args': {
        'month':"*",
        'day': "1",
        'hour': "0",
        'minute': "0",
        'second': "0",
    }
})

class AbstractSubmitter:
    """
    Abstract base class for different submitters. By design, does not use any implementation specific
    storages or submit targets. Those are defined by sub-classes and runtime provided log_facade class.
    """
    _batch_limit: int
    _chunk_size: int
    _log_facade: Type[AbstractLogFacade]

    def __init__(
            self,
            log_facade: Type[AbstractLogFacade],
            batch_limit: int = 5000,
            chunk_size: int = 500,
            submit_unsent_entries_trigger: Optional[IntervalOptions] = _default_submit_unsent_entries_trigger,
            clear_sent_entries_trigger: Optional[IntervalOptions] = _default_clear_sent_entries_trigger,
    ) -> None:
        self._log_facade = log_facade
        self._batch_limit = batch_limit
        self._chunk_size = chunk_size

        if not submit_unsent_entries_trigger:
            logger.info("NOT scheduling submit_entries, submit_unsent_entries_trigger is set to None.")

        if not clear_sent_entries_trigger:
            logger.info("NOT scheduling clear_old_entries, clear_sent_entries_trigger is set to None.")

        if (submit_unsent_entries_trigger or clear_sent_entries_trigger) and self.should_schedule():
            scheduler = BackgroundScheduler()
            scheduler.start()

            if submit_unsent_entries_trigger:
                scheduler.add_job(
                    self.submit_unsent_entries,
                    id="submit_unsent_entries",
                    name="submit_unsent_entries",
                    max_instances=1,
                    replace_existing=True,
                    trigger=submit_unsent_entries_trigger.get('trigger'),
                    **submit_unsent_entries_trigger.get('args'),
                )

            if clear_sent_entries_trigger:
                scheduler.add_job(
                    self.clear_sent_entries,
                    id="clear_sent_entries",
                    name="clear_sent_entries",
                    max_instances=1,
                    replace_existing=True,
                    trigger=clear_sent_entries_trigger.get('trigger'),
                    **clear_sent_entries_trigger.get('args'),
                )

    @abstractmethod
    def _submit_entry(self, entry: AbstractLogFacade) -> Optional[str]:
        """This method is different for each submitter, so it's required to override this."""
        return NotImplemented
    
    @transaction.atomic
    def submit_entry(self, entry: AbstractLogFacade) -> Optional[str]:
        result_id = self._submit_entry(entry)
        if result_id is not None:
            entry.mark_sent()
            return result_id
        
        return None

    def submit(self, level: int, message: Any, context: Any) -> Optional[str]:
        entry = self.get_log_facade().create(level, message, context)
        return self.submit_entry(entry)

    @transaction.atomic
    def submit_unsent_entries(self) -> Optional[List[str]]:
        result_ids: List[str] = []

        for count, entry in enumerate(self.get_unsent_entries()):
            if count >= self._batch_limit:
                print(f"Job limit of {self._batch_limit} logs reached, stopping...")
                break

            result_id = self.submit_entry(entry)

            if result_id is not None:
                entry.mark_sent()
                result_ids.append(result_id)

        return result_ids

    def get_unsent_entries(self) -> Generator[AbstractLogFacade, None, None]:
        return self.get_log_facade().get_unsent_entries(self._chunk_size)
    
    def clear_sent_entries(self, days_to_keep: int = 30) -> None:
        return self.get_log_facade().clear_sent_entries(days_to_keep)
    
    def get_log_facade(self) -> Type[AbstractLogFacade]:
        if not self._log_facade:
            raise Exception("self._log_facade is None, cannot proceed without it.")
        
        return self._log_facade

    @staticmethod
    def should_schedule():
        if os.environ.get("SUBMITTER_SCHEDULED", None) is None:
            os.environ['SUBMITTER_SCHEDULED'] = 'True'
            return True

        return False 


import logging
from typing import Optional, Type, override
from elasticsearch import Elasticsearch
from resilient_logger.abstract_log_facade import AbstractLogFacade
from resilient_logger.abstract_submitter import AbstractSubmitter

logger = logging.getLogger(__name__)

# Constants
ES_STATUS_CREATED = "created"

class ElasticsearchSubmitter(AbstractSubmitter):
    """
    Submitter class that sends the resilient log entries to Elasticsearch.
    """
    _client: Elasticsearch
    _index: str

    def __init__(self,
            log_facade: Type[AbstractLogFacade],
            client: Elasticsearch,
            index: str,
            batch_limit: int = 5000,
            chunk_size: int = 500,
    ) -> None:
        super().__init__(log_facade, batch_limit, chunk_size)

        if not client:
            raise Exception(f"ElasticsearchSubmitter is missing required argument client.")

        if not index:
            raise Exception(f"ElasticsearchSubmitter is missing required argument index.")
        
        self._index = index
        self._client = client
    
    @override
    def _submit_entry(self, entry: AbstractLogFacade) -> Optional[str]:
        document = entry.get_context()
        document["level"] = entry.get_level()
        document["message"] = entry.get_message()

        response = self._client.index(
            index=self._index,
            id=str(entry.get_id()),
            document=document,
            op_type="create",
        )

        logger.info(f"Sending status: {response}")

        if response.get("result") == ES_STATUS_CREATED:
            return response.get("_id")
        
        return None


from abc import abstractmethod
from typing import Any, Generator, Self, Union

class AbstractLogFacade:
    """
    Abstract base class (interface) that defines the method signatures.
    This is required because Django will not work if we import something that relies on Models too early.
    """
    @abstractmethod
    def get_id(self) -> Union[str, int]:
        return NotImplemented
    
    @abstractmethod
    def get_level(self) -> int:
        return NotImplemented

    @abstractmethod
    def get_message(self) -> Any:
        return NotImplemented
    
    @abstractmethod
    def get_context(self) -> Any:
        return NotImplemented
    
    @abstractmethod
    def is_sent(self) -> bool:
        return NotImplemented

    @abstractmethod
    def mark_sent(self) -> None:
        return NotImplemented
    
    @classmethod
    @abstractmethod
    def create(cls, level: int, message: Any, context: Any) -> Self:
        return NotImplemented
    
    @classmethod
    @abstractmethod
    def get_unsent_entries(cls, chunk_size: int) -> Generator[Self, None, None]:
        return NotImplemented

    @classmethod
    @abstractmethod
    def clear_sent_entries(cls, days_to_keep: int = 30) -> None:
        return NotImplemented


from importlib import import_module
import logging
from typing import Type, TypeVar

BUILTIN_LOG_RECORD_ATTRS = {
    'args',
    'asctime',
    'created',
    'exc_info',
    'exc_text',
    'filename',
    'funcName',
    'levelname',
    'levelno',
    'lineno',
    'module',
    'msecs',
    'message',
    'msg',
    'name',
    'pathname',
    'process',
    'processName',
    'relativeCreated',
    'stack_info',
    'taskName',
    'thread',
    'threadName',
}

TClass = TypeVar("TClass")

def dynamic_class(type: Type[TClass], class_path: str) -> Type[TClass]:
    """
    Loads dynamically class of given type from class_path
    and ensures it's sub-class of given input type.
    """
    parts = class_path.split('.')
    class_name = parts.pop()
    module_name = '.'.join(parts)
    module = import_module(module_name)
    cls = getattr(module, class_name)

    if not issubclass(cls, type):
        raise Exception(f"Class '{class_path}' is not sub-class of the {type}.")
    
    return cls

def get_log_record_extra(record: logging.LogRecord):
    """Returns `extra` passed to the logger."""
    return {
        name: record.__dict__[name]
        for name in record.__dict__
        if name not in BUILTIN_LOG_RECORD_ATTRS
    }


import logging
import uuid
from typing import Optional, Type, override
from resilient_logger.abstract_log_facade import AbstractLogFacade
from resilient_logger.abstract_submitter import AbstractSubmitter

class ProxySubmitter(AbstractSubmitter):
    """
    Submitter class that sends the resilient log entries to another logger.
    """
    _logger: logging.Logger

    def __init__(self,
            log_facade: Type[AbstractLogFacade],
            name: str = __name__,
            batch_limit: int = 5000,
            chunk_size: int = 500,
    ) -> None:
        super().__init__(log_facade, batch_limit, chunk_size)
        self._logger = logging.getLogger(name)
    
    @override
    def _submit_entry(self, entry: AbstractLogFacade) -> Optional[str]:
        self._logger.log(entry.get_level(), entry.get_message(), extra=entry.get_context() or {})
        return str(uuid.uuid4())